import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import processing.sound.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class GalacticSkirmish extends PApplet {

// This program runs a tower defense game
// Copyright Ryan O'Connor, David Vargas, Jacob Klukas, Lincoln Grinenko, Samuel Robey 2020



SceneGame sceneGame;
SceneTitle sceneTitle;
SceneLevelSelect sceneSelect;
SceneLoss sceneLoss;
SceneWin sceneWin;
Level level;
Enemies enemies;
Pathfinder pathfinder;

float currTime;
float prevTime = 0;
float dt;

// 0 == nothing
// 1 == shooter tower
// 2 == acid tower
// 3 = spaceship tower
int towerSelector = 0;

float mult = 1;

int levelToLoad = 1;

String priceToPay = "";

boolean rightMouse = false;
boolean leftMouse = false;
boolean prevMousePressed = false;
boolean hasRemoveTool = false;
boolean hasUpgradeTool = false;

PImage rock;//Graphic for the boulders which populate levels
PImage tower1;//Shooter tower image
PImage tower2;//Acid tower image
PImage spaceship;
PImage baseEnemy;//Image for basic enemy type
PImage tankEnemy;//Sprite for tanky enemies
PImage fastEnemy;//Sprite for fast-moving enemies
PImage flyEnemy;//Sprite for flying enemies
PImage splash;//Menu splash art
PImage levelBG;//In-game background
PImage homeBase;//Base the player guards
PImage enemyBase;
PImage acidPatch;//Graphic for acid puddles from un-upgraded towers
PImage wall;//Image for placeable wall tiles
PImage acidUpgrade;//Image for acid puddles from upgraded towers
PImage bulletUpgrade;//Image for bullets from upgraded shooter towers
PImage ufoBullet;//Image for spaceship tower bullets
PImage explosion;//Image for explosions from spaceship tower
PImage smonk;//Smoke particle
PImage bgMoon;
PImage bgMars;

float baseHealth = 100;

//Button variables:
float buttonWidth = 300;//Determines default width of buttons.
float buttonHeight = 100;//Determines default height of buttons.
float buttonRadii = 10;//Determines default roundness of button corners.
ArrayList<Button> buttons = new ArrayList();//Array list which stores all buttons.

//Sam's experimental variables - Should never be true for final build
boolean extraDeaths = true; //5% chance to proc during enemy death when enabled
boolean biggerTowers = false; //looks nifty until you place a tower above another tower

ArrayList<Particle> particles = new ArrayList();

PFont font;//Font in the style of the title screen
PFont moneyFont;//Font similar to the title screen text but with clearer numbers

PVector wind;

SoundFile hoverSFX;
SoundFile pressSFX;
SoundFile bgmSFX;
SoundFile dying;
SoundFile death01;
SoundFile death02;
SoundFile shoot;

public void setup() {
  
  TileHelper.app = this;
  pathfinder = new Pathfinder();
  rectMode(CENTER);
  switchToTitle();
  rock = loadImage("redWall.png");
  tower1 = loadImage("tower01.png");
  tower2 = loadImage("tower02.png");
  spaceship = loadImage("spaceship.png");
  baseEnemy = loadImage("baseEnemy.png");
  tankEnemy = loadImage("tankEnemy.png");
  fastEnemy = loadImage("fastEnemy.png");
  flyEnemy = loadImage("flyEnemy.png");
  splash = loadImage ("splash.png");
  levelBG = loadImage ("levelBG.png");
  homeBase= loadImage ("homeBase.png");
  enemyBase = loadImage ("enemyBase.png");
  acidPatch = loadImage ("acidPatch.png");
  acidUpgrade = loadImage ("acidUpgrade.png");
  bulletUpgrade = loadImage ("bulletUpgrade.png");
  wall = loadImage ("wall.png");
  ufoBullet = loadImage ("UFObullet.png");
  explosion = loadImage ("explosion.png");
  smonk = loadImage("smonk.png");
  bgMoon = loadImage("bgMoon.png");
  bgMars = loadImage("bgMars.png");

  font = createFont("fml.ttf", 32);
  textFont(font, 32);
  moneyFont = createFont("fyl.ttf", 32);

  hoverSFX = new SoundFile(this, "hover.wav");
  pressSFX = new SoundFile(this, "press.wav");
  //bgmSFX = new SoundFile(this, "bgm.wav");
  dying = new SoundFile(this, "dying.wav");
  death01 = new SoundFile(this, "death01.wav");
  death02 = new SoundFile(this, "death02.wav");
  shoot = new SoundFile(this, "shoot.wav");
  pressSFX.amp(0.75f);
  shoot.amp(0.15f);
  //bgmSFX.amp(0.15);
  
}

public void draw() {
  //DELTAAA TIMEEE
  dt = calcDeltaTime();
  //if (!bgmSFX.isPlaying()) bgmSFX.play();
  //Switch Scene
  if (sceneTitle != null) {
    sceneTitle.update();
    if (sceneTitle != null) sceneTitle.draw(); // this extra if statement exists because the sceneTitle.update() might result in the scene switching...
  } else if (sceneGame != null) {
    sceneGame.update();
    if (sceneGame != null) sceneGame.draw(); // this extra if statement exists because the scenePlay.update() might result in the scene switching...
  } else if (sceneSelect != null) {
    sceneSelect.update();
    if (sceneSelect != null) sceneSelect.draw();
  }else if (sceneWin != null) {
    sceneWin.update();
    if (sceneWin != null) sceneWin.draw();
  }else if (sceneLoss != null) {
    sceneLoss.update();
    if (sceneLoss != null) sceneLoss.draw();
  }
}

public void switchToTitle() {
  sceneTitle = new SceneTitle();
  sceneGame = null;
  sceneSelect = null;
  sceneWin = null;
  sceneLoss = null;
}
public void switchToGame() {
  sceneGame = new SceneGame();
  sceneTitle = null;
  sceneSelect = null;
  sceneWin = null;
  sceneLoss = null;
}
public void switchToSelect() {
  sceneSelect = new SceneLevelSelect();
  sceneGame = null;
  sceneTitle = null;
  sceneWin = null;
  sceneLoss = null;
}
public void switchToLoss() {
  sceneLoss = new SceneLoss();
  sceneSelect = null;
  sceneGame = null;
  sceneTitle = null;
  sceneWin = null;
}
public void switchToWin() {
  sceneWin = new SceneWin();
  sceneSelect = null;
  sceneGame = null;
  sceneTitle = null;
  sceneLoss = null;
}


public void mousePressed() {
  if (mouseButton == LEFT) leftMouse = true;
  if (mouseButton == RIGHT) rightMouse = true;
}
public void mouseReleased() {
  if (mouseButton == LEFT) leftMouse = false;
  if (mouseButton == RIGHT) rightMouse = false;
}

public float calcDeltaTime() {
  currTime = millis();
  float deltaTime = (currTime - prevTime)/1000;
  prevTime = currTime;
  return deltaTime;
}

public void newButton(String s, String n, float x, float y, float w, float h, float r, boolean c) {
  Button b = new Button(s, n, x, y, w, h, r, c);
  buttons.add(b);
}

//Iterates through the button array and executes the update function.
public void updateButtons() {
  for (int i = 0; i < buttons.size(); i++) {
    buttons.get(i).update();
  }
}

public void setButtonsToTrue() {
  for (int i = 0; i < buttons.size(); i++) {
    buttons.get(i).canBePressed = true;
  }
}

//Iterates through the button array and executes the draw function.
public void drawButtons() {
  for (int i = 0; i < buttons.size(); i++) {
    buttons.get(i).draw();
  }
}

public void fuckThoseButtons() {
  for (int i = buttons.size()-1; i >= 0; i--) {
    buttons.remove(i);
  }
}

public void fuckThoseParticles() {
  for (int i = particles.size()-1; i >= 0; i--) {
    particles.remove(i);
  }
}

public void burst(float x, float y, String type, float amount) {
  for (int i = 0; i < amount; i++) {
    Particle p = new Particle(x, y, type);
    particles.add(p);
  }
}

//Iterates through the particles array and executes the update function.
public void updateParticles() {
  //particles.applyForce(wind);
  for (int i = 0; i < particles.size(); i++) {
    particles.get(i).a.x = wind.x;
    particles.get(i).update();
  }
}

//Iterates through the particles array and executes the draw function.
public void drawParticles() {
  for (int i = 0; i < particles.size(); i++) {
    particles.get(i).draw();
  }
}


public void polygon(float x, float y, float radius, int npoints) {
  float angle = TWO_PI / npoints;
  beginShape();
  for (float a = 0; a < TWO_PI; a += angle) {
    float sx = x + cos(a) * radius;
    float sy = y + sin(a) * radius;
    vertex(sx, sy);
  }
  endShape(CLOSE);
}
class AcidTower extends Tower {  

  public final PVector tileLocation = new PVector(0, 0);
  int poisonTimer = 50;
  int poisonReset = poisonTimer;
  boolean hasPoisoned = false;


  AcidTower(PVector tileLocation) {
    super(tileLocation);
    radius = 75;
    baseDamage = 10;
    shootTimer = PApplet.parseInt(75 / mult);
    shootTimerReset = shootTimer;
  }

  public void draw() {
    super.draw();
    fill(0, 255, 0);

    towerType = 2;

    //circle(pixlP.x, pixlP.y, radius * 2);

    image(tower2, pixlP.x - 10, pixlP.y - 10, 20, 20);

    //circle(pixlP.x,pixlP.y,20);
  }

  public void update() {
    super.update();

    if (hasPoisoned) poisonTimer--;
    if (poisonTimer <= 0) {
      poisonTimer = poisonReset;
      hasPoisoned = false;
    }    
    if (upgradeTool == 2) {
      radius = 100;
      baseDamage = 15;
    }
    if (upgradeTool == 3) {
      radius = 125;
      baseDamage = 20;
    }
    if (upgradeTool == 4) {
      radius = 150;
    }
  }
}
class Bullet {

  float x;
  float y;
  float angle;
  float speed = 500;
  float radius = 10;
  boolean isDead = false;
  int damage;
  int towerType;
  int towerUpgrade;
  boolean canDamageFlying = false;

  Bullet(float posX, float posY, float angleT, int baseDamage, int tType, int uType) {
    x = posX;
    y = posY;
    angle = angleT;
    damage = baseDamage;
    towerType = tType;
    towerUpgrade = uType;
    shoot.play();
  }

  public void update() {
    x += cos(angle) * speed * dt;
    y += sin(angle) * speed * dt;
    if (x < -100) isDead = true;
    if (x > width + 100) isDead = true;
    if (y < 160) isDead = true;
    if (y > height + 100) isDead = true;
  }

  public void draw() {
    pushMatrix();
    fill(0);
    noStroke();
    if (towerType == 1) {
      if (towerUpgrade < 4) {
        fill(100, 100, 255);
        circle(x - 5, y - 5, 10);
      } else if (towerUpgrade >= 4) {
        image(bulletUpgrade, x - 5, y - 5, 10, 10);
      }
    }
    if (towerType == 2) {      
      if (towerUpgrade < 4) {
        image(acidPatch, x - 5, y - 5, 10, 10);
      } else if (towerUpgrade >= 4) {
        image(acidUpgrade, x - 5, y - 5, 10, 10);
      }
    }
    if (towerType == 3) {      
      image(ufoBullet, x - 5, y - 5, 10, 10);
      canDamageFlying = true;
    }

    popMatrix();
  }
}
class Button {
  float inc = 2.5f;//increment by which stats increase
  float x;      //x position
  float y;      //y position
  float w;      //width
  float h;      //height
  float r;      //roundness of corners
  String name;  //name of button
  String desc;  //button tag
  int relPos;   //relative x position: 1, 2, and 3 are equivalent to left, middle and right respectively

  Boolean deleteThisShit = false;
  Boolean sfxReady = false;
  Boolean canBePressed = true;


  Button(String tag, String title, float xpos, float ypos, float buttonWidth, float buttonHeight, float radii, boolean isActive) {
    x = xpos;
    y = ypos;
    w = buttonWidth;
    h = buttonHeight;
    r = radii;
    name = title;
    desc = tag;
    if (x<width/2) relPos = 1;
    else if (x > 2*width/5 && x < 3*width/5) relPos = 2;
    else relPos = 3;
    canBePressed = isActive;
  }

  public void update() {
    //if(sfxReady) hoverSFX.play();
    if (isHovered() && leftMouse && !prevMousePressed && canBePressed) press();
  }

  public void draw() {
    stroke(255);
    strokeWeight(4);       
    if (isHovered()) fill(237, 28, 36);    
    else fill(0);

    if (hasRemoveTool && desc == "Remove") fill(237, 28, 36);
    if (hasUpgradeTool && desc == "Upgrade") fill(237, 28, 36);



    if (towerSelector == 1 && desc == "Tower1") { 
      priceToPay = "-$5";
      name = "";
      fill(237, 28, 36);
    } else if (towerSelector != 1 && desc == "Tower1") {
      name = "$5";
    }

    if (towerSelector == 2 && desc == "Tower2") { 
      priceToPay = "-$10";
      name = "";
      fill(237, 28, 36);
    } else if (towerSelector != 2 && desc == "Tower2") {
      name = "$10";
    }

    if (towerSelector == 3 && desc == "Tower3") { 
      priceToPay = "-$15";
      name = "";
      fill(237, 28, 36);
    } else if (towerSelector != 3 && desc == "Tower3") {
      name = "$15";
    }

    if (towerSelector == 4 && desc == "Wall") {
      priceToPay = "-$5";
      name = "";
      fill(237, 28, 36);
    } else if (towerSelector != 1 && desc == "Wall") {
      name = "Wall:\n$5";
    }

    if (!canBePressed) {
      fill(50);
    }

    rect(x, y, w, h, r);

    if (desc == "Tower1") textFont(moneyFont, 24);
    if (desc == "Tower2") textFont(moneyFont, 24);
    if (desc == "Tower3") textFont(moneyFont, 24);


    fill(255);
    textAlign(CENTER, CENTER);
    textSize(48);
    text(name, x, y);
  }

  public Boolean isHovered() {
    float left = x - w/2;
    float right = x + w/2;
    float top = y - h/2;
    float bot = y + h/2;

    if (mouseX >= left && mouseX <= right && mouseY >= top && mouseY <= bot && sfxReady) {
      hoverSFX.play();
      sfxReady = false;
      return true;
    } else if (mouseX >= left && mouseX <= right && mouseY >= top && mouseY <= bot && !sfxReady) {
      return true;
    } else {
      sfxReady = true;
      return false;
    }
  }

  public void press() {
    pressSFX.play();
    println("The " + name + " button is pressed!!");
    switch(desc) {
    case "Play":
      fuckThoseButtons();
      fuckThoseParticles();
      switchToSelect();
      break;
    case "Quit":
      if (sceneGame == null) {
        exit();
      } else {
        fuckThoseButtons();
        fuckThoseParticles();
        switchToTitle();
        break;
      }
    case "1":
      //change level to 1 in place of this line
      fuckThoseButtons();
      fuckThoseParticles();
      levelToLoad = 1;
      level = new Level();
      switchToGame();
      break;
    case "2":
      //change level to 2 in place of this line
      fuckThoseButtons();
      fuckThoseParticles();
      levelToLoad = 2;
      level = new Level();
      switchToGame();
      break;
    case "3":
      //change level to 3 in place of this line
      fuckThoseButtons();
      fuckThoseParticles();
      levelToLoad = 3;
      level = new Level();
      switchToGame();
      break;
    case "4":
      //change level to 4 in place of this line
      fuckThoseButtons();
      fuckThoseParticles();
      levelToLoad = 4;
      level = new Level();
      switchToGame();
      break;
    case "5":
      //change level to 5 in place of this line
      fuckThoseButtons();
      fuckThoseParticles();
      levelToLoad = 5;
      level = new Level();
      switchToGame();
      break;
    case "Upgrade":
      if (hasUpgradeTool) hasUpgradeTool = false;
      else hasUpgradeTool = true;
      hasRemoveTool = false;
      towerSelector = 0;
      break;
    case "Remove":
      if (hasRemoveTool) hasRemoveTool = false;
      else hasRemoveTool = true;
      hasUpgradeTool = false;
      towerSelector = 0;
      break;
    case "Back":
      fuckThoseButtons();
      fuckThoseParticles();
      switchToTitle();
      break;
    case "Tower1":
      towerSelector = 1;
      hasUpgradeTool = false;
      hasRemoveTool = false;
      break;
    case "Tower2":
      towerSelector = 2;
      hasUpgradeTool = false;
      hasRemoveTool = false;
      break;
    case "Tower3":
      towerSelector = 3;
      hasUpgradeTool = false;
      hasRemoveTool = false;
      break;
    case "Wall":
      towerSelector = 4;
      hasUpgradeTool = false;
      hasRemoveTool = false;
      break;
    case "Restart":
      fuckThoseButtons();
      fuckThoseParticles();      
      level = new Level();
      switchToGame();
      break;
    }
  }
}
class Emitter {
  float x;
  float y;
  float lifetime;
  String type;
  float frequency;
  Boolean isDead = false;
  float timer;
  Emitter(float xPos, float yPos, String ptype, float duration, float particlesPerSecond) {
    lifetime = duration;
    x = xPos;
    y = yPos;
    type = ptype;
    frequency = particlesPerSecond;
    timer = 0;
  }
  public void update() {
    timer+=dt;
    lifetime-=dt;
    if (lifetime<0) isDead = true;
    else if (!isDead) {
      if (timer >= 1/frequency) {
        burst(x-10, y-10, type, 1);
        timer -= 1/frequency;
      }
      //freq
    }
  }
}
class Enemies {

  // GRID-SPACE COORDINATES:
  Point gridP = new Point();
  Point gridT = new Point();

  // PIXEL-SPACE COORDINATES:
  PVector pixlP = new PVector(); // CURRENT PIXEL POSITION
  PVector pixlT = new PVector(); 

  PImage enemy;

  float x;
  float y;

  Tile tile = level.enemySpawn;
  Wall wall;

  ArrayList<Tile> path;
  boolean findPath = false;
  boolean canAttack = false;
  float radius = 16;
  float angle;
  float prevAngle;
  int attackTimer = 100;
  int health = 100;
  boolean isDead = false;
  boolean isDying = false;
  boolean isFlying = false;
  boolean isSlowed = false;
  int slowTimer = 0;
  int slowTimerReset = 50;
  float speed = .1f;
  float speedReset;
  float dmg = 5;

  //Determines what enemy is spawned
  //1 = basic
  //2 = fast
  //3 = tank
  //4 = flying
  int enemyType = 1;

  //variables to change based on which enemy spawns
  //Health
  //speed
  //isFlying

  Enemies(Point spawnP, Point targetP, int eType) {
    teleportTo(spawnP);
    setTargetPosition(targetP);
    enemyType = eType;
    setEnemyValues();
  }

  public void teleportTo(Point gridP) {
    Tile tile = level.getTile(gridP);
    if (tile != null) {
      this.gridP = gridP.get();
      this.gridT = gridP.get();
      this.pixlP = tile.getCenter();
    }
  }

  public void setTargetPosition(Point gridT) {
    this.gridT = gridT.get();
  }

  public void update() {
    if (!isDying) {
      if (findPath == true) findPathAndTakeNextStep();
      updateMove();
      attackTimer--;
      x = X;
      y = Y;
      if (attackTimer < 0) attackTimer = 0;

      //stop moving, stay and attack wall until wall is gone

      if (canAttack == true) {
        wall.health--;
      }
      if (isSlowed) speed = speedReset * .5f;
      if (!isSlowed) speed = speedReset;
      if (slowTimer <= 0) {
        isSlowed = false;
        slowTimer = 0;
      }
      slowTimer--;
      if (health <= 0) {
        if (!extraDeaths) isDead = true;
        else {
          if (random(0, 1) > 0.95f) {
            isDying = true;
            dying.play();
          } else isDead = true;
        }
      }
    } else {
      angle+=5*PI*dt;
      if (!dying.isPlaying()) {
        isDead = true;
      }
    }
  }

  public void findPathAndTakeNextStep() {
    findPath = false;
    // FIND PATH TO TARGET GRID POSITION
    Tile start = level.getTile(gridP);
    Tile end = level.getTile(gridT);
    if (start == end) {
      path = null;
      isDead = true;
      baseHealth-=10;
      return;
    }

    // CHANGE CURRENT GRID POSITION TO NEXT STEP TOWARDS TARGET
    path = pathfinder.findPath(start, end);

    if (path != null && path.size() > 1) {
      tile = path.get(1);

      if (tile.isPassable()) {
        gridP = new Point(tile.X, tile.Y);
        canAttack = false;
      }
    }
  }

  public void updateMove() {
    angleToNext();
    float snapThreshold = 1;


    // GET THE TARGET POSITION IN PIXELS
    pixlT = level.getTileCenterAt(gridP);
    PVector diff = PVector.sub(pixlT, pixlP);

    // MOVE CURRENT PIXEL POSITION TOWARDS TARGET
    pixlP.x += diff.x * speed;
    pixlP.y += diff.y * speed;

    // ENSURE NO MOVEMENT BUGS
    if (abs(diff.x) < snapThreshold) pixlP.x = pixlT.x;
    if (abs(diff.y) < snapThreshold) pixlP.y = pixlT.y;

    // IF WE'VE ARRIVED AT LOCATION, TRIGGER NEXT PATH FIND
    if (pixlT.x == pixlP.x && pixlT.y == pixlP.y) findPath = true;
  }

  public void draw() {
    noStroke();


    pushMatrix();
    translate( pixlP.x, pixlP.y);    
    if (angle != 0) rotate(angle - degrees(40));
    if (angle == 0) rotate(prevAngle - degrees(40));
    image(enemy, - 10, - 10, 20, 20);

    popMatrix();

    //fill(0);
    //ellipse(pixlP.x, pixlP.y, 16, 16);
    //drawPath();
  }

  public void drawPath() {
    // DRAW PATH
    if (path != null && path.size() > 1) {
      stroke(0);
      PVector prevP = pixlP.get();
      for (int i = 1; i < path.size(); i++) {
        PVector currP = path.get(i).getCenter();
        strokeWeight(2);
        line(prevP.x, prevP.y, currP.x, currP.y);
        prevP = currP;
      }
      noStroke();
      ellipse(prevP.x, prevP.y, 8, 8);
    }
  }


  public void angleToNext() {
    prevAngle = angle;
    float dx = pixlP.x - pixlT.x;
    float dy = pixlP.y - pixlT.y;
    angle =  atan2(dy, dx);
  }

  public void setEnemyValues() {
    if (enemyType == 1) {
      health = 100;
      speed = .1f * mult;
      isFlying = false;  
      enemy = baseEnemy;
    }
    if (enemyType == 2) {
      health = 60;
      speed = .2f * mult;
      isFlying = false;  
      enemy = fastEnemy;
    }
    if (enemyType == 3) {
      health = 350;
      speed = .07f * mult;
      isFlying = false;  
      enemy = tankEnemy;
    }
    if (enemyType == 4) {
      health = 150;
      speed = .11f * mult;
      isFlying = true;  
      enemy = flyEnemy;
    }
    speedReset = speed;
  }
}
class Level {

  int[][] level;
  Tile[][] tiles;
  boolean useDiagonals = false;

  Tile enemySpawn;
  Tile enemyGoal;

  Level() {
    if (levelToLoad == 1) {
      loadLevel(LevelDefs.LEVEL1);
      mult = 1;
    } else if (levelToLoad == 2) {
      loadLevel(LevelDefs.LEVEL2);
      mult = 1;
    } else if (levelToLoad == 3) {
      loadLevel(LevelDefs.LEVEL3);
      mult = 1;
    } else if (levelToLoad == 4) {
      loadLevel(LevelDefs.LEVEL4);
      mult = 1.3f;
    } else {
      loadLevel(LevelDefs.LEVEL5);
      mult = 1.8f;
    }
  }

  public void draw() {
    noStroke();
    for (int Y = 0; Y < tiles.length; Y++) {
      for (int X = 0; X < tiles[Y].length; X++) {
        tiles[Y][X].draw();
        if (tiles[Y][X].TERRAIN == 3) enemyGoal = tiles[Y][X];
        if (tiles[Y][X].TERRAIN == 6) enemySpawn = tiles[Y][X];
      }
    }

    fill(0);
  }

  // return the tile at grid position (X, Y)
  public Tile getTile(int X, int Y) {
    // TODO: implement method. check for potential run-time errors.
    if (X < 0 || Y < 0) return null;
    if (Y >= tiles.length || X >= tiles[0].length) return null;
    return tiles[Y][X];
  }

  // convenience method
  public Tile getTile(Point p) {
    return getTile(p.x, p.y);
  }

  // return the pixel coordinates of the center
  // of the tile at specified grid position
  public PVector getTileCenterAt(Point p) {
    // TODO: implement method. check for potential run-time errors.
    Tile tile = getTile(p);
    if (tile == null) return new PVector();
    return tile.getCenter();
  } 

  // returns true if the tile at the specified
  // grid position is passable
  public boolean isPassable(Point p) {
    // TODO: implement method. check for potential run-time errors.
    Tile tile = getTile(p);
    if (tile == null) return false;
    return tile.isPassable();
  }

  // reload the level. useful for re-establishing
  // neighbor relationships between the tiles
  public void reloadLevel() {
    loadLevel(level);
  }

  public void loadLevel(int[][] layout) {

    level = layout; // cache the layout (to enable reloading levels)

    // build our multidimensional array of tiles:
    int ROWS = layout.length;
    int COLS = layout[0].length;
    tiles = new Tile[ROWS][COLS];
    for (int Y = 0; Y < ROWS; Y++) {
      for (int X = 0; X < COLS; X++) {
        Tile tile = new Tile(X, Y);
        tile.TERRAIN = layout[Y][X];
        tiles[Y][X] = tile;
      }
    }
    // TODO: set each tile's neighboring nodes
    for (int Y = 0; Y < ROWS; Y++) {
      for (int X = 0; X < COLS; X++) {
        if (TileHelper.isHex) {
          if (X % 2 == 0) {
            tiles[Y][X].addNeighbors(new Tile[] {
              getTile(X - 1, Y), 
              getTile(X - 1, Y + 1), 
              getTile(X, Y - 1), 
              getTile(X, Y + 1), 
              getTile(X + 1, Y), 
              getTile(X + 1, Y + 1)
              });
          } else {
            tiles[Y][X].addNeighbors(new Tile[] {
              getTile(X - 1, Y - 1), 
              getTile(X - 1, Y), 
              getTile(X, Y - 1), 
              getTile(X, Y + 1), 
              getTile(X + 1, Y - 1), 
              getTile(X + 1, Y)
              });
          }
        } else {
          tiles[Y][X].addNeighbors(new Tile[] {
            getTile(X - 1, Y), 
            getTile(X + 1, Y), 
            getTile(X, Y - 1), 
            getTile(X, Y + 1), 
            });
          if (useDiagonals) {
            tiles[Y][X].addNeighbors(new Tile[] {
              getTile(X - 1, Y - 1), 
              getTile(X + 1, Y + 1), 
              getTile(X + 1, Y - 1), 
              getTile(X - 1, Y + 1), 
              });
          } //end if diagonals
        } //end else for isHex
      } //end X for loop
    } //end Y for loop
  } //end loadLevel()

  // toggle the use of diagonals
  public void toggleDiagonals() {
    useDiagonals = !useDiagonals;
    reloadLevel();
  }
}
static class LevelDefs {

  // THE MOVEMENT COSTS (CORRESPONDS TO THE TERRAIN TYPE IN LEVELS)
  // For example, terrain 0 has a moveCost of 1, while terrin 1,
  // has a moveCost of 4, and terrain 2 has a moveCost of 1000.
  static  int[] moveCost = {1, 4, 1000, 1, 1, 20, 1};
  //0 = basic terrain
  //1 = player placed towers
  //2 = impassable terrain
  //3 = middle of base
  //4 = outskirts of base, can't place (also used for blocking the player from placing a tower on the enemy spawn
  //5 = player placed walls
  //6 = enemy spawn (THIS NEEDS TO HAVE A CIRCLE WITH 7 TILES ON EACH SIDE OF EITHER TERRAIN 2 OR TERRAIN 4
  //7 = poison tiles
  

  // THESE SHOULDN'T BE JAGGED!
  static int[][] LEVEL1 = ///////////////////LINCOLN LEVEL
    {
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 2, 2, 4, 4, 4, 4, 4, 4, 4, 2, 2, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0}, 
    {0, 2, 2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0}, 
    {2, 2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0}, 
    {2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 6, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 4, 4, 4}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 4, 3, 4}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 4, 4, 4}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {2, 2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {0, 2, 2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {0, 0, 2, 2, 4, 4, 4, 4, 4, 4, 4, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0}, 
    {0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
  };
  static int[][] LEVEL2 = ///////////////////SAM LEVEL
    {
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2}, 
    {4, 6, 4, 4, 4, 4, 4, 4, 4, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0}, 
    {4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2}, 
    {4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2}, 
    {0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2}, 
    {2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 2, 2}, 
    {2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 2, 2, 0, 2}, 
    {0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 2, 2, 0, 2, 2, 2}, 
    {2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 2, 2, 2, 0, 2}, 
    {2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 2, 2, 2}, 
    {0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 2}, 
    {2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2}, 
    {2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2}, 
    {0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 4, 0}, 
    {0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 2, 2, 0, 2, 2, 0, 0, 0, 0, 4, 3, 4}, 
    {2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 4, 4, 4} 
    
  };
  static int[][] LEVEL3 = /////////////////// SAM LEVEL 2 ELECTRIC BOOGALOO
    {
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 4, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 3, 4}, 
    {4, 6, 4, 4, 4, 4, 4, 4, 4, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 2, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 0, 2, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 0, 2, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 0, 2, 2, 2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 4, 0, 4, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0}, 
    {2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0}, 
    
  };
  static int[][] LEVEL4 = /////////////////// DAVIDS LEVEL
    {
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 4, 4, 4, 0, 0}, 
    {0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 4, 3, 4, 0, 0}, 
    {0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 4, 4, 4, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 6, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2},  
     
  };
  static int[][] LEVEL5 = /////////////////// JACOBS LEVEL
    {
    {2, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0}, 
    {0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0}, 
    {0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0}, 
    {2, 0, 0, 0, 2, 0, 2, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0}, 
    {2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0}, 
    {0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2}, 
    {0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2}, 
    {0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2}, 
    {0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2}, 
    {0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2}, 
    {0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 2, 2, 2, 2, 2, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2}, 
    {2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2}, 
    {2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2}, 
    {2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2}, 
    {2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2}, 
    {2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2}, 
    {2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 0, 2, 2, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 2, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 6, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 4, 3, 4, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 4, 4, 4, 0, 0}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0},   
    
  };
  static int[][] LEVEL6 = /////////////////// LEVEL
    {
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 4, 4, 4, 0, 0}, 
    {0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 4, 3, 4, 0, 0}, 
    {0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 4, 4, 4, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0}, 
    {0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 6, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {0, 0, 0, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    {2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2}, 
    
  };
  /*static int[][] LEVEL6 =
   new int[64][30];
   void makeNoise(float noise) {
   int assignVal;
   for (int i = 0; i < LEVEL6.length; i++) {
   for (int j = 0; j < LEVEL6[0].length; j++) {
   float rand = norm(noise/i + noise/j, 0, 2);
   int fin = (int)map(rand, 0, 2, 0, 5);
   if (fin <= 2) {
   assignVal = fin;
   } else {
   assignVal = 1000;
   }
   LEVEL6[i][j] = assignVal;
   }
   }
   }*/
}
class Particle {
  float x;
  float y;
  String type;
  PVector v = new PVector();
  PVector a = new PVector();
  float minSpeed = -100;
  float maxSpeed = 100;
  float decay = 1;
  float maxVel = 550;
  Boolean xVelGreaterThanZero; //Whether x velocity is greater than zero
  Boolean yVelGreaterThanZero; //Whether y velocity is greater than zero
  float lifespan = 128;
  Particle(float xPos, float yPos, String name) {
    x = xPos;
    y = yPos;
    type = name;
    switch (type) {
    case "dust":
      v.x = random(minSpeed, maxSpeed);
      v.y = random(minSpeed, maxSpeed);
      break;
    case "wind":
      x = random(0, width);
      y = random(150, height);
      v.x = 0;
      v.y = 0;
      break;
    case "wind2":
      x = random(0, width);
      y = random(150, height);
      v.x = 0;
      v.y = 0;
      maxVel = maxVel-50;
      break;
    case "wind3":
      x = random(0, width);
      y = random(150, height);
      v.x = 0;
      v.y = 0;
      maxVel = maxVel-100;
      break;
    case "wind4":
      x = random(0, width);
      y = random(150, height);
      v.x = 0;
      v.y = 0;
      maxVel = maxVel-150;
      break;    
    case "smonk":
      v.x = random(-50, 50);
      v.y = random(-100, -200);
      maxVel = maxVel-125;
      break;
    }

    if (v.x > 0) xVelGreaterThanZero = true;
    else xVelGreaterThanZero = false;
    if (v.y > 0) yVelGreaterThanZero = true;
    else yVelGreaterThanZero = false;
  }
  public void update() {
    switch (type) {
    case "dust":
      lifespan -= 50*dt;
      if(yVelGreaterThanZero) v.y-=decay;
      else v.y+=decay;
      if (yVelGreaterThanZero && v.y < 0) v.y = 0;
      else if (!yVelGreaterThanZero && v.y >0) v.y = 0;
      break;
    case "wind":
      if (x > width + 10) x = -5;
      if (x < -10) x = width + 5;
      break;
    case "wind2":
      if (x > width + 10) x = -5;
      if (x < -10) x = width + 5;
      break;
    case "wind3":
      if (x > width + 10) x = -5;
      if (x < -10) x = width + 5;
      break;
    case "wind4":
      if (x > width + 10) x = -5;
      if (x < -10) x = width + 5;
      break;
    case "smonk":
      lifespan -= 250*dt;
      break;
    }

    v.x+=a.x*2.5f;
    if (v.x > maxVel) v.x = maxVel;
    else if (v.x < -maxVel) v.x = -maxVel;
    if (v.y > maxVel) v.y = maxVel;
    else if (v.y < -maxVel) v.y = -maxVel;
    x+=v.x*dt;
    y+=v.y*dt;


    //if(xVelGreaterThanZero) v.x-=decay;
    //else v.x+=decay;


    //if(xVelGreaterThanZero && v.x < 0) v.x = 0;
    //else if(!xVelGreaterThanZero && v.x >0) v.x = 0;
    //if(yVelGreaterThanZero && v.y < 0) v.y = 0;
    //else if(!yVelGreaterThanZero && v.y >0) v.y = 0;
  }
  public void draw() {

    switch (type) {
    case "dust":
      tint(255, lifespan);
      noStroke();
      fill(255, 0, 0);
      //ellipse(x, y, 4, 4);
      image(rock, x, y, 3, 3);
      tint(255);
      break;
    case "wind":
      tint(255, 128);
      noStroke();
      fill(255, 0, 0);
      //ellipse(x, y, 4, 4);
      image(rock, x, y, 3, 3);
      tint(255);
      break;
    case "wind2":
      tint(255, 128);
      noStroke();
      fill(255, 0, 0);
      //ellipse(x, y, 4, 4);
      image(rock, x, y, 3, 3);
      tint(255);
      break;
    case "wind3":
      tint(255, 128);
      noStroke();
      fill(255, 0, 0);
      //ellipse(x, y, 4, 4);
      image(rock, x, y, 3, 3);
      tint(255);
      break;
    case "wind4":
      tint(255, 128);
      noStroke();
      fill(255, 0, 0);
      //ellipse(x, y, 4, 4);
      image(rock, x, y, 3, 3);
      tint(255);
      break;
    case "smonk":
      tint(255, lifespan);
      noStroke();
      fill(255, 0, 0);
      //ellipse(x, y, 4, 4);
      image(smonk, x, y);
      tint(255);
      break;
    }
  }
}

class Pathfinder {

  boolean useManhattan = false;
  boolean useDiagonals = false;
  ArrayList<Tile> open = new ArrayList<Tile>(); // collection of tiles we can use to solve the algorithm
  ArrayList<Tile> closed = new ArrayList<Tile>(); // collection of tiles that we've already checked and should not check again

  Pathfinder() {
  }

  public ArrayList<Tile> findPath(Tile start, Tile end) {
    open.clear(); // empty array
    closed.clear(); // empty array

    // TODO: starting tile's "parent" property should be null
    start.resetParent();

    // STEP 1: CONNECT START TO END
    connectStartToEnd(start, end);

    // STEP 2: BUILD PATH BACK TO BEGINNING
    ArrayList<Tile> path = new ArrayList<Tile>();
    Tile pathNode = end;
    while (pathNode != null) {
      path.add(pathNode);
      pathNode = pathNode.parent;
    }

    // STEP 3: REVERSE THE PATH (FOR CONVENIENCE)
    ArrayList<Tile> rev = new ArrayList<Tile>();
    int maxIndex = path.size() - 1;
    for (int i = maxIndex; i >= 0; i--) {
      rev.add(path.get(i));
    }

    return rev;
  }

  // CONNECT THE START TO THE END
  public void connectStartToEnd(Tile start, Tile end) {

    open.add(start); // ADD start TILE TO open ARRAY

    // WHILE WE STILL HAVE TILES IN THE OPEN ARRAY
    while (open.size () > 0) {

      // FIND THE OPEN TILE WITH THE LOWEST F VALUE
      float F = 9999;
      int index = -1;

      for (int i = 0; i < open.size(); i++) {
        Tile temp = open.get(i);
        if (temp.F < F) {
          F = temp.F;
          index = i;
        }
      }

      // REMOVE THE TILE FROM THE OPEN ARRAY
      Tile current = open.remove(index);

      // ADD THE TILE FROM THE CLOSED ARRAY
      closed.add(current);

      // IF WE'VE CONNECTED THE START TO THE END, BREAK OUT OF WHILE, RETURN OUT OF FUNCTION
      if (current == end) {
        // Path was found!
        break;
      }

      // LOOP THROUGH ALL OF current's NEIGHBORS
      for (int i = 0; i < current.neighbors.size(); i++) {
        Tile neighbor = current.neighbors.get(i);
        //// IF THE NEIGHBOR ISN'T IN CLOSED ARRAY:
        if (!tileInArray(closed, neighbor)) {
          //// IF THE NEIGHBOR ISN'T IN OPEN ARRAY:
          if (!tileInArray(open, neighbor)) {
            ////// ADD IT TO THE OPEN ARRAY
            open.add(neighbor);
            ////// SET IT'S PARENT
            neighbor.setParent(current);
            ////// CALCULATE IT'S H VALUE
            neighbor.toggleHeuristics(end, useManhattan);
          } else {      //// ELSE
            ////// IF IT WOULD BE CHEAPER TO MOVE TO THAT TILE FROM current
            if (neighbor.G > current.G + neighbor.getTerrainCost()) {
              //////// SET IT'S PARENT TO current
              neighbor.setParent(current);
              //////// RECALCULATE IT'S H VALUE
              neighbor.toggleHeuristics(end, useManhattan);
            } // end if
          } //end else for !tileInArray
        } // end if
      } // end for loop
    } // end while loop
  } // end method

  // returns true if the specified tile is contained
  // within the specified ArrayList
  public boolean tileInArray(ArrayList<Tile> a, Tile t) {
    // TODO: implement the method
    for (int i = 0; i < a.size(); i++) {
      if (a.get(i) == t) return true;
    }
    return false;
  }

  // toggle the heuristic that we're using
  public void toggleHeuristic() {
    useManhattan = !useManhattan;
  }
}
class Point {
  int x = 0;
  int y = 0;
  Point() {
  }
  Point(int x, int y) {
    this.x = x;
    this.y = y;
  }
  public Point get() {
    return new Point(x, y);
  }
  public Point getUp() {
    return new Point(x, y-1);
  }
  public Point getDown() {
    return new Point(x, y+1);
  }
  public Point getLeft() {
    return new Point(x-1, y);
  }
  public Point getRight() {
    return new Point(x+1, y);
  }
  public String toString() {
    return x + ", " + y;
  }
}
class SceneGame {

  Point g;
  Tile tile;

  ArrayList<Enemies> enemies = new ArrayList();
  ArrayList<Tower> towers = new ArrayList();
  ArrayList<Bullet> bullets = new ArrayList();
  ArrayList<Wall> walls = new ArrayList();
  ArrayList<Emitter> emitters = new ArrayList();

  boolean doOnce = false;
  boolean paused = false;

  int currency = 15;
  int currencyAmount = 15;  

  //when the player beats wave 10 beats the game, open a win screen or main menu
  int wave = 0;
  boolean isInWave = false;
  boolean isSpawningEnemies = false;
  int waveTimer = 500;
  int spawnTimer = 0;
  int basicEnemiesToSpawn = 0;
  int fastEnemiesToSpawn = 0;
  int tankEnemiesToSpawn = 0;
  int flyingEnemiesToSpawn = 0;
  int totalEnemiesToSpawn = 0;
  int totalEnemiesSpawned = 0;
  int spawnValue = 1;
  int lowSpawnValue;
  int highSpawnValue;
  
  PVector m;

  float windSpeed = 20;
  Boolean add = false;

  float windParticles = 200; //This MUST be evenly divisible by 4 otherwise you will get an error. Bigger numbers look nicer but drop framerate


  //int phase;
  //1: Placement phase
  //
  //

  //when buttons are added, if remove tool is selected change hasupgradetool to false and vice versa. also set towerSelector = 0 so they can't place any towers. when tower is selected, set tower selector to appropiate tower #

  SceneGame() {
    baseHealth = 100;
    // Prevents Double Clicking on Load
    leftMouse = false;
    rightMouse = false;
    prevMousePressed = false;
    //Spawn Buttons
    newButton("Upgrade", "Upgrade", width/1.275f, height/18, 240, 50, buttonRadii, true);
    newButton("Remove", "Remove", width/1.275f, height/6.65f, 240, 50, buttonRadii, true);
    newButton("Quit", "Quit", width - 74.5f, 74.5f, 100, 100, buttonRadii, true);
    newButton("Tower1", "$5", width/2 - 300, 74.5f, 100, 100, buttonRadii, true);
    newButton("Tower2", "$10", width/2 - 180, 74.5f, 100, 100, buttonRadii, false);
    newButton("Tower3", "$15", width/2 - 60, 74.5f, 100, 100, buttonRadii, false);
    newButton("Wall", "Wall:\n$5", width/2 + 120, 74.5f, 200, 100, buttonRadii, true);



    for (int i = 0; i < (windParticles/4); i++) {
      Particle p = new Particle(0, 0, "wind");
      particles.add(p);
    }
    for (int i = 0; i < (windParticles/4); i++) {
      Particle p = new Particle(0, 0, "wind2");
      particles.add(p);
    }
    for (int i = 0; i < (windParticles/4); i++) {
      Particle p = new Particle(0, 0, "wind3");
      particles.add(p);
    }
    for (int i = 0; i < (windParticles/4); i++) {
      Particle p = new Particle(0, 0, "wind4");
      particles.add(p);
    }
  }

  public void update() { 

    if (baseHealth<=0) {
      fuckThoseButtons();
      fuckThoseParticles();
      switchToLoss();
    }

    if (wave >= 4) setButtonsToTrue();

    //WIND
    //float dx = map(mouseX, 0, width, -20, 20); //Maps wind to mouse for debugging
    if (windSpeed >= 20) add = false;
    else if (windSpeed <= -20) add = true;

    if (add)windSpeed+=dt;
    else windSpeed-=dt;

    if (windSpeed>20)windSpeed = 20;
    else if (windSpeed<-20) windSpeed = -20;

    wind = new PVector(windSpeed, 0);
    //println(wind.x);


    //background(198, 123, 92);
    //image(bgMoon, -5, 145, width, height-145);
    image(bgMoon, -5, 145);

    // Draw Level & Buttons
    level.draw();
    updateEmitters();
    updateWaves();



    // Gets the Tile Mouse is Hovering
    g = TileHelper.pixelToGrid(new PVector(mouseX, mouseY - 150));
    tile = level.getTile(g);
    if (tile != null) { 
      tile.isHovered = true;
      m = tile.getCenter();
      fill(0);
      //ellipse(m.x, m.y, 8, 8);
    }    

    // Updates Enemy Array
    for (int i = 0; i < enemies.size(); i++) {
      enemies.get(i).update();
      if (enemies.get(i).pixlP.x == enemies.get(i).gridT.x && enemies.get(i).pixlP.y == enemies.get(i).gridT.y) {
        enemies.get(i).isDead = true;
        baseHealth-=enemies.get(i).dmg;
      }
      enemies.get(i).draw();  
      for (int j = 0; j < walls.size(); j++) {
        if (enemies.get(i).tile.getCenter().x == walls.get(j).pixlP.x && enemies.get(i).tile.getCenter().y == walls.get(j).pixlP.y) {
          enemies.get(i).wall = walls.get(j);
          enemies.get(i).canAttack = true;
          //if(enemies.get(i).wall == null) enemies.get(i).tile.TERRAIN = 0;
        }
      }
      if (enemies.get(i).isDead) {
        if (random(0, 1) > 1f) death01.play();
        else death02.play();
        burst(enemies.get(i).pixlP.x, enemies.get(i).pixlP.y, "dust", 25);
        enemies.remove(i);
        totalEnemiesSpawned--;
        currency++;
      }
    }

    //Updates walls array
    try {
      for (int k = 0; k < walls.size(); k++) {
        walls.get(k).update();
        walls.get(k).draw();
        if (walls.get(k).isDead) {       
          for (int i = 0; i < enemies.size(); i++) {
            if (enemies.get(i).tile.getCenter().x == walls.get(k).pixlP.x && enemies.get(i).tile.getCenter().y == walls.get(k).pixlP.y) {
              enemies.get(i).tile.TERRAIN = 0;
              enemies.get(i).tile.hasATower = false;
              spawnSmokeEmitter(walls.get(k).pixlP.x, walls.get(k).pixlP.y);
              walls.remove(k);
            }
          }
        }
      }
    } 
    catch (IndexOutOfBoundsException e) {
    }

    // Updates Tower Array
    for (int i = 0; i < towers.size(); i++) {
      towers.get(i).update();
      towers.get(i).draw();

      // Gets all Enemies
      // If Tower and Enemy Are Close and ShootTimer <= 0
      // Spawn a Bullet
      for (int j = 0; j < enemies.size(); j++) {
        if (towers.get(i).distanceTowerEnemy(towers.get(i), enemies.get(j)) && towers.get(i).shootTimer <= 0 ) {
          if (towers.get(i).canHitFlying) {
            Bullet b = new Bullet(towers.get(i).pixlP.x, towers.get(i).pixlP.y, towers.get(i).angleToEnemy(towers.get(i), enemies.get(j)), towers.get(i).baseDamage, towers.get(i).towerType, towers.get(i).upgradeTool);
            bullets.add(b);
            towers.get(i).shootTimer = towers.get(i).shootTimerReset;
          } else if (!towers.get(i).canHitFlying && !enemies.get(j).isFlying) {
            Bullet b = new Bullet(towers.get(i).pixlP.x, towers.get(i).pixlP.y, towers.get(i).angleToEnemy(towers.get(i), enemies.get(j)), towers.get(i).baseDamage, towers.get(i).towerType, towers.get(i).upgradeTool);
            bullets.add(b);
            towers.get(i).shootTimer = towers.get(i).shootTimerReset;
          }
        }
        if (towers.get(i).distanceToExplosionRadius(towers.get(i), enemies.get(j)) && towers.get(i).upgradeTool == 4 && towers.get(i).towerType == 3 && towers.get(i).explosionTimer <= 0) {          
          enemies.get(j).health -= 110;
          spawnSmokeEmitter(enemies.get(j).pixlP.x, enemies.get(j).pixlP.y);
          towers.get(i).hasExploded = true;
        }
      }




      // If hovering a tile that is not null, and has upgrade tool
      // Change priceToPay variable depending on the type of tower and it's upgrade level

      // Else if has removal tool
      // Change priceToPay variable depending on the type of tower  


      if (hasUpgradeTool && tile != null && tile.isHovered) {
        println(priceToPay);
        if (tile.getCenter().x == towers.get(i).pixlP.x && tile.getCenter().y == towers.get(i).pixlP.y) {
          if (towers.get(i).upgradeTool == 1) {
            if (towers.get(i).towerType == 1) {              
              priceToPay = "-$10";
              noFill();
              stroke(255, 255, 255, 100);        
              circle(tile.getCenter().x, tile.getCenter().y, 350);
            } else if (towers.get(i).towerType == 2) {
              priceToPay = "-$15";
              noFill();
              stroke(255, 255, 255, 100);        
              circle(tile.getCenter().x, tile.getCenter().y, 200);
            } else if (towers.get(i).towerType == 3) {
              priceToPay = "-$20";
              noFill();
              stroke(255, 255, 255, 100);        
              circle(tile.getCenter().x, tile.getCenter().y, 400);
            }
          }  
          if (towers.get(i).upgradeTool == 2) {
            if (towers.get(i).towerType == 1) {
              priceToPay = "-$15";
              noFill();
              stroke(255, 255, 255, 100);        
              circle(tile.getCenter().x, tile.getCenter().y, 400);
            } else if (towers.get(i).towerType == 2) {
              priceToPay = "-$20";
              noFill();
              stroke(255, 255, 255, 100);        
              circle(tile.getCenter().x, tile.getCenter().y, 250);
            } else if (towers.get(i).towerType == 3) {
              priceToPay = "-$25";
              noFill();
              stroke(255, 255, 255, 100);        
              circle(tile.getCenter().x, tile.getCenter().y, 450);
            }
          }  
          if (towers.get(i).upgradeTool == 3) {
            if (towers.get(i).towerType == 1) {
              priceToPay = "-$20";
              noFill();
              stroke(255, 255, 255, 100);        
              circle(tile.getCenter().x, tile.getCenter().y, 450);
            } else if (towers.get(i).towerType == 2) {
              priceToPay = "-$25";
              noFill();
              stroke(255, 255, 255, 100);        
              circle(tile.getCenter().x, tile.getCenter().y, 300);
            } else if (towers.get(i).towerType == 3) {
              priceToPay = "-$30";
              noFill();
              stroke(255, 255, 255, 100);        
              circle(tile.getCenter().x, tile.getCenter().y, 500);
            }
          }  
          if (towers.get(i).upgradeTool == 4) {
            priceToPay = "MAX";
          }
        } else priceToPay = "";
      } else if (hasRemoveTool && tile != null && tile.isHovered) {
        if (tile.getCenter().x == towers.get(i).pixlP.x && tile.getCenter().y == towers.get(i).pixlP.y) {
          if (towers.get(i).towerType == 1) {
            priceToPay = "+$0";
          } else if (towers.get(i).towerType == 2) {
            priceToPay = "+$5";
          } else if (towers.get(i).towerType == 3) {
            priceToPay = "+$10";
          }
        } else priceToPay = "";
      } else priceToPay = "";
    }

    // Updates Bullets
    try {
      for (int i = bullets.size()-1; i >= 0; i--) {
        bullets.get(i).update();
        bullets.get(i).draw();
        for (int j = enemies.size()-1; j >= 0; j--) {
          if (detectCollision(bullets.get(i), enemies.get(j))) { 
            if (enemies.get(j).isFlying && bullets.get(i).canDamageFlying) {
              enemies.get(j).health -= 30;
              if (bullets.get(i).towerType == 2 && bullets.get(i).towerUpgrade == 4) {
                enemies.get(j).isSlowed = true;
                enemies.get(j).slowTimer = enemies.get(j).slowTimerReset;
              }
              bullets.remove(i);
            } else if (enemies.get(j).isFlying && !bullets.get(i).canDamageFlying) {
              return;
            } else if (!enemies.get(j).isFlying) {
              enemies.get(j).health -= 30;
              if (bullets.get(i).towerType == 2 && bullets.get(i).towerUpgrade == 4) {
                enemies.get(j).isSlowed = true;
                enemies.get(j).slowTimer = enemies.get(j).slowTimerReset;
              }
              bullets.remove(i);
            }
          }
        }
        if (bullets.get(i).isDead == true) bullets.remove(i);
      }
    } 
    catch (IndexOutOfBoundsException e) {
    }

    // On Left Click
    if (leftMouse && !prevMousePressed) {
      // Check if there is a tile
      if (tile != null) {
        // Check if tile has a tower, have a tower selected, don't have remove tool, and don't have upgrade tool
        if (!tile.hasATower && towerSelector != 0 && !hasRemoveTool && !hasUpgradeTool && tile.TERRAIN != 2 && tile.TERRAIN != 3 && tile.TERRAIN != 4 && tile.TERRAIN != 6) { 
          // Spawn towers based on towerSelector and if you have enough currency
          if (towerSelector == 1 && currency >= 5) {
            PVector tileLocation = tile.getCenter();
            ShooterTower t = new ShooterTower(tileLocation);
            towers.add(t);
            towerSelector = 0;
            currency -= 5;
            //priceToPay = "";
            tile.TERRAIN = 1;
            tile.hasATower = true;
          } else if (towerSelector == 2 && currency >= 10) {
            PVector tileLocation = tile.getCenter();
            AcidTower t = new AcidTower(tileLocation);
            towers.add(t);
            towerSelector = 0;
            currency -= 10;
            //priceToPay = "";
            tile.TERRAIN = 1;
            tile.hasATower = true;
          } else if (towerSelector == 3 && currency >= 15) {
            PVector tileLocation = tile.getCenter();
            SpaceshipTower t = new SpaceshipTower(tileLocation);
            towers.add(t);
            towerSelector = 0;
            currency -= 15;
            //priceToPay = "";
            tile.TERRAIN = 1;
            tile.hasATower = true;
          } else if (towerSelector == 4 && currency >=5) {
            PVector tileLocation = tile.getCenter();
            Wall w = new Wall(tileLocation);
            walls.add(w);
            currency -= 5;
            //priceToPay = "";
            tile.TERRAIN = 5;
            tile.hasATower = true;
          } else {
            towerSelector = 0;
            priceToPay = "";
          }
        } else if (hasRemoveTool) { // If you have the remove tool
          // for all towers
          for (int i = 0; i < towers.size(); i++) {
            // compare tile center to tower x and y              
            if (tile.getCenter().x == towers.get(i).pixlP.x && tile.getCenter().y == towers.get(i).pixlP.y) {    
              // delete if equal to each other 
              if (towers.get(i).towerType == 1) {
                currency += 0;
              } else if (towers.get(i).towerType == 2) {
                currency += 5;
              } else if (towers.get(i).towerType == 3) {
                currency += 10;
              }
              towers.remove(i);
              tile.TERRAIN = 0;
              hasRemoveTool = false;
            }
          }
          for (int i = 0; i < walls.size(); i++) {
            if (tile.getCenter().x == walls.get(i).pixlP.x && tile.getCenter().y == walls.get(i).pixlP.y) {
              //spawnSmokeEmitter(walls.get(i).pixlP.x, walls.get(i).pixlP.y);
              walls.remove(i);
              tile.TERRAIN = 0;
            }
          }
          if (!tile.hasATower) hasRemoveTool = false;
        } else if (hasUpgradeTool) { // If you have the upgrade tool
          // for all towers
          for (int i = 0; i < towers.size(); i++) {
            // compare tile center to tower x and y   
            if (tile.getCenter().x == towers.get(i).pixlP.x && tile.getCenter().y == towers.get(i).pixlP.y) {
              // upgarde if equal to each one
              if (towers.get(i).upgradeTool == 1) {
                if (towers.get(i).towerType == 1 && currency >= 10) {
                  currency -= 10;
                  towers.get(i).upgradeTool++;
                } else if (towers.get(i).towerType == 2 && currency >= 15) {
                  currency -= 15;
                  towers.get(i).upgradeTool++;
                } else if (towers.get(i).towerType == 3 && currency >= 20) {
                  currency -= 20;
                  towers.get(i).upgradeTool++;
                }
              } else if (towers.get(i).upgradeTool == 2 && currency >= 15) {
                if (towers.get(i).towerType == 1) {
                  currency -= 15;
                  towers.get(i).upgradeTool++;
                } else if (towers.get(i).towerType == 2 && currency >= 20) {
                  currency -= 20;
                  towers.get(i).upgradeTool++;
                } else if (towers.get(i).towerType == 3 && currency >= 25) {
                  currency -= 25;
                  towers.get(i).upgradeTool++;
                }
              } else if (towers.get(i).upgradeTool == 3) {
                if (towers.get(i).towerType == 1 && currency >= 20) {
                  currency -= 20;
                  towers.get(i).upgradeTool++;
                } else if (towers.get(i).towerType == 2 && currency >= 25) {
                  currency -= 25;
                  towers.get(i).upgradeTool++;
                } else if (towers.get(i).towerType == 3 && currency >= 30) {
                  currency -= 30;
                  towers.get(i).upgradeTool++;
                }
              } else if (towers.get(i).upgradeTool >= 4) {
                hasUpgradeTool = false;
              }
            }
          }
          if (!tile.hasATower) hasUpgradeTool = false;
        } else {
          towerSelector = 0;
        }

        // set target position for enemy on click
        for (int i = 0; i < enemies.size(); i++) {
          //enemies.get(i).setTargetPosition(TileHelper.pixelToGrid(new PVector(mouseX, mouseY - 150)));
        }
      }
    }

    // If nothing selected priceToPay is empty
    if (!hasUpgradeTool && !hasRemoveTool && towerSelector == 0) priceToPay = "";

    // On right click remove anything selected
    if (rightMouse && !prevMousePressed) {
      towerSelector = 0;
      hasUpgradeTool = false;
      hasRemoveTool = false;
    }

    // Draws Towers on Mouse Cursor if Selected
    if (towerSelector == 1) {
      if (tile != null) {
        noStroke();
        fill(255, 0, 0);
        image(tower1, tile.getCenter().x - 7.5f, tile.getCenter().y - 7.5f, 15, 15);
        noFill();
        stroke(255, 255, 255, 100);        
        circle(tile.getCenter().x, tile.getCenter().y, 300);
      } else {
        noStroke();
        fill(255, 0, 0);
        circle(mouseX, mouseY, 15);
      }
    } else if (towerSelector == 2) {
      if (tile != null) {
        noStroke();
        fill(0, 255, 0);
        image(tower2, tile.getCenter().x - 7.5f, tile.getCenter().y - 7.5f, 15, 15);
        noFill();
        stroke(255, 255, 255, 100);        
        circle(tile.getCenter().x, tile.getCenter().y, 175);
      } else {
        noStroke();
        fill(0, 255, 0);
        circle(mouseX, mouseY, 15);
      }
    } else if (towerSelector == 3) {
      if (tile != null) {
        noStroke();
        fill(0, 0, 255);
        image(spaceship, tile.getCenter().x - 7.5f, tile.getCenter().y - 7.5f, 15, 15);
        noFill();
        stroke(255, 255, 255, 100);        
        circle(tile.getCenter().x, tile.getCenter().y, 400);
      } else {
        noStroke();
        fill(0, 0, 255);
        circle(mouseX, mouseY, 15);
      }
    }

    // Animates currency shown on screen
    if (currencyAmount > currency) currencyAmount--;
    else if (currencyAmount < currency) currencyAmount++;
    else currencyAmount = currency;

    //    if (roundTimer <= 0) {
    //      startRound();
    //      roundTimer = 30;
    //    } else {
    //      roundTimer--;
    //    }
    updateParticles();
    drawParticles();
    fill(150, 28, 36);
    noStroke();
    rect(width/2, 74.5f, width, 149);    
    strokeWeight(2);
    stroke(0);
    line(0, 149, width, 149);
    line(0, 74.5f, width, 74.5f);
    drawButtons();
    updateButtons();
    fill(255);
    textFont(moneyFont, 24);
    text("Wave " + wave, width/20, 120.5f);
    textFont(moneyFont, 18);
    text("Remaining Enemies:  " + totalEnemiesSpawned, width/10.2f, 15.5f);
    text("Base Health: " + floor(baseHealth), 100, 140.5f); 

    // Draw Text Near Mouse
    textFont(moneyFont, 54);
    text("$" + currencyAmount, width/8, 73.5f);
    textFont(moneyFont, 24);
    text(priceToPay, mouseX + 20, mouseY - 20);
    textFont(font, 32);

    noFill();
    strokeWeight(2);
    stroke(0);
    polygon(m.x, m.y, 12, 6);    

    // NO DOUBLE CLICKING ALLOWED
    prevMousePressed = rightMouse;
    prevMousePressed = leftMouse;
  }

  public void draw() {
  }

  public boolean detectCollision(Bullet a, Enemies b) {
    float dx = b.pixlP.x - a.x; 
    float dy = b.pixlP.y - a.y; 
    float dis = sqrt(dx * dx + dy * dy);
    if (dis <=a.radius +b.radius) return true;  
    return false;
  }

  public void spawnEnemies() {
    if (basicEnemiesToSpawn > 0 && spawnValue == 1) {
      Enemies e = new Enemies(TileHelper.pixelToGrid(TileHelper.gridToPixel(level.enemySpawn.X, level.enemySpawn.Y)), TileHelper.pixelToGrid(TileHelper.gridToPixel(level.enemyGoal.X, level.enemyGoal.Y)), 1);
      enemies.add(e);
      totalEnemiesToSpawn--;
      basicEnemiesToSpawn--;
      spawnTimer = PApplet.parseInt(random(lowSpawnValue, highSpawnValue));
      return;
    }
    if (fastEnemiesToSpawn > 0 && spawnValue == 2) {
      Enemies e = new Enemies(TileHelper.pixelToGrid(TileHelper.gridToPixel(level.enemySpawn.X, level.enemySpawn.Y)), TileHelper.pixelToGrid(TileHelper.gridToPixel(level.enemyGoal.X, level.enemyGoal.Y)), 2);
      enemies.add(e);
      totalEnemiesToSpawn--;
      fastEnemiesToSpawn--;
      spawnTimer = PApplet.parseInt(random(lowSpawnValue, highSpawnValue));
      return;
    }
    if (tankEnemiesToSpawn > 0 && spawnValue == 3) {
      Enemies e = new Enemies(TileHelper.pixelToGrid(TileHelper.gridToPixel(level.enemySpawn.X, level.enemySpawn.Y)), TileHelper.pixelToGrid(TileHelper.gridToPixel(level.enemyGoal.X, level.enemyGoal.Y)), 3);
      enemies.add(e);
      totalEnemiesToSpawn--;
      tankEnemiesToSpawn--;
      spawnTimer = PApplet.parseInt(random(lowSpawnValue, highSpawnValue));
      return;
    }
    if (flyingEnemiesToSpawn > 0 && spawnValue == 4) {
      Enemies e = new Enemies(TileHelper.pixelToGrid(TileHelper.gridToPixel(level.enemySpawn.X, level.enemySpawn.Y)), TileHelper.pixelToGrid(TileHelper.gridToPixel(level.enemyGoal.X, level.enemyGoal.Y)), 4);
      enemies.add(e);
      totalEnemiesToSpawn--;
      flyingEnemiesToSpawn--;
      spawnTimer = PApplet.parseInt(random(lowSpawnValue, highSpawnValue));
      return;
    }
    return;
  }

  public void updateWaves() {    
    if (!isInWave) {
      waveTimer--;
      if (waveTimer <= 0) {
        isInWave = true;
        wave++;
      }
    }    
    if (isInWave) {
      if (!isSpawningEnemies) {
        if (wave == 1) {         
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(40 / mult);
          highSpawnValue = PApplet.parseInt(100/ mult);
          basicEnemiesToSpawn = 10;
          fastEnemiesToSpawn = 0;
          tankEnemiesToSpawn = 0;
          flyingEnemiesToSpawn = 0;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
        } else if (wave == 2) {
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(40/ mult);
          highSpawnValue = PApplet.parseInt(95/ mult);
          basicEnemiesToSpawn = 20;
          fastEnemiesToSpawn = 0;
          tankEnemiesToSpawn = 0;
          flyingEnemiesToSpawn = 0;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
        } else if (wave == 3) {
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(40/ mult);
          highSpawnValue = PApplet.parseInt(90/ mult);
          basicEnemiesToSpawn = 20;
          fastEnemiesToSpawn = 5;
          tankEnemiesToSpawn = 0;
          flyingEnemiesToSpawn = 0;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
        } else if (wave == 4) {
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(40/ mult);
          highSpawnValue = PApplet.parseInt(85/ mult);
          basicEnemiesToSpawn = 20;
          fastEnemiesToSpawn = 10;
          tankEnemiesToSpawn = 0;
          flyingEnemiesToSpawn = 0;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
          setButtonsToTrue();
          try {
            for (int k = buttons.size(); k >= 0; k--) {
              if (buttons.get(k).name == "Tower1" || buttons.get(k).name == "Tower2") {
                buttons.get(k).canBePressed = true;
              }
            }
          }
          catch(IndexOutOfBoundsException e) {
          }
        } else if (wave == 5) {
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(35/ mult);
          highSpawnValue = PApplet.parseInt(80/ mult);
          basicEnemiesToSpawn = 20;
          fastEnemiesToSpawn = 10;
          tankEnemiesToSpawn = 10;
          flyingEnemiesToSpawn = 5;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
        } else if (wave == 6) {
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(30/ mult);
          highSpawnValue = PApplet.parseInt(75/ mult);
          basicEnemiesToSpawn = 25;
          fastEnemiesToSpawn = 15;
          tankEnemiesToSpawn = 10;
          flyingEnemiesToSpawn = 10;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
        } else if (wave == 7) {
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(25/ mult);
          highSpawnValue = PApplet.parseInt(70/ mult);
          basicEnemiesToSpawn = 15;
          fastEnemiesToSpawn = 40;
          tankEnemiesToSpawn = 5;
          flyingEnemiesToSpawn = 5;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
        } else if (wave == 8) {
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(15/ mult);
          highSpawnValue = PApplet.parseInt(60/ mult);
          basicEnemiesToSpawn = 15;
          fastEnemiesToSpawn = 15;
          tankEnemiesToSpawn = 30;
          flyingEnemiesToSpawn = 5;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
        } else if (wave == 9) {
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(5/ mult);
          highSpawnValue = PApplet.parseInt(40/ mult);
          basicEnemiesToSpawn = 15;
          fastEnemiesToSpawn = 15;
          tankEnemiesToSpawn = 10;
          flyingEnemiesToSpawn = 40;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
        } else if (wave == 10) {
          // set the amount of enemies to spawn this wave
          lowSpawnValue = PApplet.parseInt(2/ mult);
          highSpawnValue = PApplet.parseInt(20/ mult);
          basicEnemiesToSpawn = 30;
          fastEnemiesToSpawn = 40;
          tankEnemiesToSpawn = 20;
          flyingEnemiesToSpawn = 40;
          totalEnemiesToSpawn = basicEnemiesToSpawn + fastEnemiesToSpawn + tankEnemiesToSpawn + flyingEnemiesToSpawn;
        } else if (wave == 11) {//EXIT THE GAME
          fuckThoseButtons();
          fuckThoseParticles();
          switchToWin();
        }
        totalEnemiesSpawned = totalEnemiesToSpawn;
        isSpawningEnemies = true;
      } else if (isSpawningEnemies && totalEnemiesToSpawn > 0) {
        spawnValue = PApplet.parseInt(random(0, 5));
        if (spawnTimer <= 0) {

          spawnEnemies();
        }
        spawnTimer--;
      }
      if (enemies.size() <= 0 && totalEnemiesToSpawn <= 0 && isSpawningEnemies) {
        isSpawningEnemies = false;
        isInWave = false;
        waveTimer = PApplet.parseInt(500 / mult);
      }
    }
  }

  public void spawnSmokeEmitter(float x, float y) {
    Emitter e = new Emitter(x, y, "smonk", 3, 32);
    emitters.add(e);
  }

  public void updateEmitters() {
    for (int i = 0; i < emitters.size(); i++) {
      emitters.get(i).update();
      if (emitters.get(i).isDead) emitters.remove(i);
    }
  }
}
class SceneLevelSelect {

  boolean hoverBack = false;
  boolean hoverOne = false;

  float levelSize = 100;

  SceneLevelSelect() {
    leftMouse = false;
    rightMouse = false;
    prevMousePressed = false;
    newButton("1", "1", width/2 - 300, height/2, levelSize, levelSize, buttonRadii, true);
    newButton("2", "2", width/2 - 150, height/2, levelSize, levelSize, buttonRadii, true);
    newButton("3", "3", width/2, height/2, levelSize, levelSize, buttonRadii, true);
    newButton("4", "4", width/2 + 150, height/2, levelSize, levelSize, buttonRadii, true);
    newButton("5", "5", width/2 + 300, height/2, levelSize, levelSize, buttonRadii, true);
    newButton("Back", "Back", width/2, height/1.25f, buttonWidth, buttonHeight, buttonRadii, true);
    baseHealth = 100;
  }

  public void draw() {
    drawButtons();
  }

  public void update() {
    background(198, 123, 92);


  image(levelBG, 0, 0 );

    // TOP TEXT
    pushMatrix();
    textSize(89);
    fill(255, 0, 0);
    text("Level Select", width/2, height/5);
    popMatrix();

    updateButtons();
  }
}
class SceneLoss {
  
  int levelToLoad;
  int fillColorR;
  int fillColorG = 85;
  int fillColorB = 170;
  boolean fillcolorRed;
  boolean fillcolorGreen;
  boolean fillcolorBlue;

  SceneLoss() {
    newButton("Back", "Back", width/2, height/1.25f, buttonWidth, buttonHeight, buttonRadii, true);
    newButton("Restart", "Restart", width/2, height/1.8f, buttonWidth, buttonHeight, buttonRadii, true);
  }

  public void draw() {
    image(bgMars, 0, 0 );
    
    if(fillColorR >= 255) fillcolorRed = true;
    if(fillColorG >= 255) fillcolorGreen = true;
    if(fillColorB >= 255) fillcolorBlue = true;
    if(fillColorR <= 0) fillcolorRed = false;
    if(fillColorG <= 0) fillcolorGreen = false;
    if(fillColorB <= 0) fillcolorBlue = false;
    if(!fillcolorRed) fillColorR++;
    if(!fillcolorGreen) fillColorG++;
    if(!fillcolorBlue) fillColorB++;
    if(fillcolorRed) fillColorR--;
    if(fillcolorGreen) fillColorG--;
    if(fillcolorBlue) fillColorB--;

    // TOP TEXT
    pushMatrix();
    textSize(89);
    fill(fillColorR, fillColorG, fillColorB);
    text("Take this fat L", width/2, height/5);
    popMatrix();
    updateButtons();
    drawButtons();
  }

  public void update() {
  }
}
class SceneTitle {

  PImage Title;
 
  //boolean hoverPlay = false;
  //boolean hoverQuit = false;

  SceneTitle() {
    leftMouse = false;
    rightMouse = false;
    prevMousePressed = false;
    //hoverPlay = false;
    //hoverQuit = false;
    Title = loadImage("GalacticSkirmishStacked.png");
  

    newButton("Play", "Play", width/6, height/1.7f, buttonWidth, buttonHeight, buttonRadii, true);
    newButton("Quit", "Quit", width/6, height/1.2f, buttonWidth, buttonHeight, buttonRadii, true);
  }

  public void draw() {
    drawButtons();
  }

  public void update() {
    // BACKGROUND
    background(198, 123, 92);

    // SPLASH IMAGE
    
    image(splash, 0, 0, 1280, 720 );


    // TITLE IMAGE
    pushMatrix();
    scale(.2f);
    image(Title, width - 1000, 221, 2744, 1326);
    popMatrix();
    
    

    // BUTTONS
    updateButtons();

    ////SPLASH ART
    //pushMatrix();
    //stroke(0);
    //strokeWeight(4);
    //fill(198, 123, 92, 0);
    //rect(width/1.35, height/2, 500, 700);
    //fill(0);
    //text("Splash Art", width/1.35, height/2);
    //popMatrix();

    //PREVENTS DOUBLE CLICKING
    prevMousePressed = leftMouse;
  }
}
class SceneWin {

  SceneWin() {
    newButton("Back", "Back", width/2, height/1.25f, buttonWidth, buttonHeight, buttonRadii, true);
    newButton("Restart", "Restart", width/2, height/1.8f, buttonWidth, buttonHeight, buttonRadii, true);
  }

  public void draw() {
    image(levelBG, 0, 0 );

    // TOP TEXT
    pushMatrix();
    textSize(89);
    fill(255, 0, 0);
    text("Victory", width/2, height/5);
    popMatrix();
    updateButtons();
    drawButtons();
  }

  public void update() {
  }
}
class ShooterTower extends Tower {  

  public final PVector tileLocation = new PVector(0, 0);

  ShooterTower(PVector tileLocation) {
    super(tileLocation);
    shootTimer = PApplet.parseInt(80 / mult);
    shootTimerReset = shootTimer;
  }

  public void draw() {
    super.draw();

    towerType = 1;

    pushMatrix();
    if(!biggerTowers) image(tower1, pixlP.x - 10 , pixlP.y - 10 , 20 , 20);
    else image(tower1, pixlP.x - 12.5f , pixlP.y - 12.5f , 25 , 25);
    //fill(255, 0, 0);
    //if (upgradeTool == 2) fill(255, 0, 255);
    //if (upgradeTool == 3) fill(255, 255, 0);
    //if (upgradeTool == 4) fill(255, 255, 255);
    //translate(pixlP.x, pixlP.y);
    //rotate(angle);
    //circle(0, 0, 20);
    popMatrix();
  }

  public void update() {
    super.update();
    
    if(upgradeTool == 2){
      radius = 175;
      baseDamage = 35;
    }
    if(upgradeTool == 3){
      radius = 200;
      baseDamage = 40;
    }
    if(upgradeTool == 4){
      radius = 225;
      shootTimerReset = PApplet.parseInt(24 / mult);
    }
  }
}
class SpaceshipTower extends Tower{  
  
  public final PVector tileLocation = new PVector(0,0);
  
  
  SpaceshipTower(PVector tileLocation){
    super(tileLocation);
    radius = 200;
    shootTimer = PApplet.parseInt(90 / mult);
    shootTimerReset = shootTimer;
    baseDamage = 40;
    canHitFlying = true;
  }
  
  public void draw(){
    super.draw();
    fill(0,0,255);
    //circle(pixlP.x,pixlP.y,explosionRadius * 2);
    towerType = 3;
    if(upgradeTool == 4 && explosionDrawTimer > 0) image(explosion, pixlP.x - 75 , pixlP.y - 75 , 150 , 150);
    
    image(spaceship, pixlP.x - 10 , pixlP.y - 10 , 20 , 20);
    
    
    
  }
  
  public void update(){
    super.update();
    explosionTimer--;
    explosionDrawTimer--;
    if(explosionTimer <= -1) explosionTimer = -1;
    if(explosionDrawTimer <= -1) explosionDrawTimer = -1;
    
    if (hasExploded) {
      explosionTimer = 800;
      explosionDrawTimer = 40;
      hasExploded = false;
    }
    
    if(upgradeTool == 2){
      radius = 200;
      baseDamage = 50;
    }
    if(upgradeTool == 3){
      radius = 225;
      shootTimerReset = PApplet.parseInt(60 / mult);
      
    }
    if(upgradeTool == 4){
      radius = 250;
      baseDamage = 60;
    }
  }
}
static class TileHelper {

  static GalacticSkirmish app;
  static boolean isHex = true;
  final static int W = 20;
  final static int H = 20;
  final static int halfW = W / 2;
  final static int halfH = H / 2;

  //PImage img;  


  // converts from grid coordinates to pixel coordinates
  public static PVector gridToPixel(Point p) {
    // TODO: implement
    return gridToPixel(p.x, p.y);
  }

  public static PVector gridToPixel(int X, int Y) {
    PVector v = new PVector(X * W, Y * H);
    if (isHex && X % 2 == 0)v.y += 10;
    return v;
  }

  // converts from pixel coordinates to grid coordinates
  public static Point pixelToGrid(PVector p) {
    // TODO: implement
    int x = (int)(p.x / W);
    if (isHex && x % 2 == 0) p.y -= halfH;
    int y = (int)(p.y / H);
    return app.new Point(x, y);
  }
}

class Tile {

  int X;
  int Y;
  int TERRAIN = 0;
  boolean isHovered = false;
  boolean hasATower = false;


  Tile(int X, int Y) {
    this.X = X;
    this.Y = Y;
  }

  public void draw() {
    // TODO: draw rectangle
    //Fills Color of Shape Based on Terrain Type
    pushMatrix();
    stroke(0);
    strokeWeight(1.5f);
    if (TERRAIN == 0) {
      if (TileHelper.isHex) fill(198, 123, 92);
      else return;
    }
    if (TERRAIN == 2) {
      PVector p = getCenter(); 
      if (hasATower == false) {
        image(rock, p.x - 10, p.y - 10, 20, 20);
      }
    }

    if ( TERRAIN == 3) {
      PVector p = getCenter();
      image(homeBase, p.x - 25, p.y - 25, 45, 45);
    }
    if (TERRAIN == 5) {
    }
    if (TERRAIN == 6) {
      PVector p = getCenter();
      image(enemyBase, p.x - 25, p.y - 25, 45, 45);
      noFill();
      stroke(255, 255, 255, 100);        
      circle(getCenter().x, getCenter().y, 300);
    }
    //if (isHovered) fill(255, 255, 0);

    //Draw Circles
    if (TileHelper.isHex) {
      PVector p = getCenter();
      //ellipse(p.x, p.y, TileHelper.W, TileHelper.H);
      noStroke();
      //fill(255);
      //polygon(p.x, p.y, (TileHelper.W/2), 6);
    } else {//Draw Rectangles
      PVector p = TileHelper.gridToPixel(X, Y);
      rect(p.x, p.y, TileHelper.W, TileHelper.H);
    }
    popMatrix();

    isHovered = false;
  }

  // returns the center of the tile in pixel coordinates
  public PVector getCenter() {
    // TODO: implement method
    PVector p = TileHelper.gridToPixel(new Point(X, Y));
    p.x += TileHelper.halfW;
    p.y += TileHelper.halfH + 150;
    return p;
  }

  public Boolean isDestructable() {
    return (TERRAIN != 5);
  }

  public Boolean isPassable() {
    if (TERRAIN == 2) return false;
    else if (TERRAIN == 5) return false;
    else return true;
  }

  ///////////////////////////////////////////////////
  /////////////// PATHFINDING STUFF: ////////////////
  ///////////////////////////////////////////////////

  // TODO: add pathfinding properties
  ArrayList<Tile> neighbors = new ArrayList<Tile>(); // LIST OF NEIGHBORING TILES
  Tile parent; // PARENT TILE (previous tile in the path)
  float G; // COST TO TRAVEL TO TILE
  float F; // TOTAL COST OF THIS TILE

  // TODO: add pathfinding methods
  // ADD THE SUPPLIED TILES TO THIS TILE'S LIST OF NEIGHBORS
  public void addNeighbors(Tile[] tiles) {
    for (Tile t : tiles) {
      if ( t != null) neighbors.add(t);
    }
  }

  // SET PARENT OF THIS TILE
  // This is used during path finding to remember where the path originated
  public void setParent(Tile n) {
    parent = n;
    G = parent.G + getTerrainCost();
  }

  // RESET THE PARENT OF THIS TILE
  // This is used when starting pathfinding. When reversing back through the path, we know we've arrived once we reach a tile without a parent tile
  public void resetParent() {
    parent = null;
    G = 0;
    F = 0;
  }

  // RETURN THE COST OF MOVING INTO THIS TILE
  // This is used during pathfinding to find the cost of the tile it's moving to
  public float getTerrainCost() {
    if (TERRAIN >= 0 && TERRAIN < LevelDefs.moveCost.length) return LevelDefs.moveCost[TERRAIN];
    return 0;
  }

  // MANHATTAN HEURISTIC CALCULATION
  public float distanceManhattan(Tile n) {
    return abs(n.X -X) + abs(n.Y - Y);
  }

  // EUCLIDEAN HEURISTIC CALCULATION
  public float distanceEuclidean(Tile n) {
    if (n != null) {
      float dx = n.X - X;
      float dy = n.Y - Y;
      return sqrt(dx * dx + dy * dy);
    } else return 0;
  }

  // TOGGLE HEURISTIC USED
  public void toggleHeuristics(Tile n, boolean useManhattan) {
    if (useManhattan) F = G + distanceManhattan(n);
    if (!useManhattan) F = G + distanceEuclidean(n);
  }
}
public class Tower { 

  boolean hasBeenPlaced = false;
  float radius = 150;
  float angle;
  float shootTimer;
  float shootTimerReset;
  int baseDamage = 30;
  
  int explosionTimer = 0;
  int explosionDrawTimer = 0;
  int explosionRadius = 100;
  boolean hasExploded = false;
  
  boolean canHitFlying = false;
  
  //PImage tower1; 

  // 1 == base level
  // 2 == damage upgrade
  // 3 == damage upgrade
  // 4 == change in type
  int upgradeTool = 1;

  int towerType;

  // PIXEL-SPACE COORDINATES:
  PVector pixlP = new PVector(); // CURRENT PIXEL POSITION

  Tower(PVector tileLocation) {
    pixlP = tileLocation;
    
  }

  public void draw() {
    
    
    
  }

  public void update() {    
    if (shootTimer > 0) {
          shootTimer--;
        } else shootTimer = 0;
  }

  public boolean distanceTowerEnemy(Tower a, Enemies b) {
    float dx = b.pixlP.x - a.pixlP.x;
    float dy = b.pixlP.y - a.pixlP.y;
    float dis = sqrt(dx * dx + dy * dy);
    if (dis < a.radius + b. radius) return true;
    return false;
  }

  public float angleToEnemy(Tower a, Enemies b) {
    float dx = b.pixlP.x - a.pixlP.x;
    float dy = b.pixlP.y - a.pixlP.y;
    return atan2(dy, dx);
  }
  
  public boolean distanceToExplosionRadius(Tower a, Enemies b) {
    float dx = b.pixlP.x - a.pixlP.x;
    float dy = b.pixlP.y - a.pixlP.y;
    float dis = sqrt(dx * dx + dy * dy);
    if (dis < a.explosionRadius + b. radius) return true;
    return false;
  }
}
public class Wall {

  PVector pixlP = new PVector(); // CURRENT PIXEL POSITION

  float x;
  float y;
  int health = 50;
  boolean isDead = false;

  Wall(PVector tileLocation) {
    pixlP = tileLocation;
    x = tileLocation.x;
    y = tileLocation.y;
  }

  public void draw() {

    image(wall, x - 10, y - 10, 20, 20);
    
    //ellipse(x - 7.5, y - 7.5, 15, 15);
  }

  public void update() {
    if(health <= 0) isDead = true;
  }
}
  public void settings() {  size(1280, 720); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "GalacticSkirmish" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
